#! /usr/local/bin/perl -w

# Utility routines to support RTFM distributions
#
# Nevil Brownlee, CAIDA / University of Auckland
#
# 1155, Tue 22 Jan 02 (NZST)

$^W = 1;  # Warnings on
use strict;

my $ln10;  # Constant for log trasforms

my (  # Globals for unpack_format_record()
   $nf,            # Nbr of attributes
   %format,        # Index in record for each attribute
   @type,          # 0 = scalar, 1 = tcpdata, 2 = distribution
   @dif);          # 1 = dif (to be summed), 0 = fixed
 
my (  # Globals for get_attrib_values()
   @attrb,     # Attribute name
   @attst,     # Test (== equal, !=, # not equal)
   @attvl,     # Lowest allowed value
   @attvh,     # Highest allowed value
   @att_str,   # True if $attvl is a string
   @nvalues,   # Nbr of allowed values in range
   $caret,     # Integer parameter value introduced by ^
   $uscore,    # String parameter value introduced by _
   $at,        # String parameter value introduced by @
   $minus,     # String parameter value introduced by -
   $shriek,    # String parameter value introduced by !
   $query);    # String parameter value introduced by ?

my %month;  # Jan..Dec => 1..12

package rt_dist;


use Time::Local;

my $last_c_time;

BEGIN {
   %main::month = (
      "Jan" =>  1, "Feb" =>  2, "Mar" =>  3, "Apr" =>  4,
      "May" =>  5, "Jun" =>  6, "Jul" =>  7, "Aug" =>  8,
      "Sep" =>  9, "Oct" => 10, "Nov" => 11, "Dec" => 12,
      );

   $last_c_time = 0;

   $main::ln10 = 2.3025850929940457;
   }  # BEGIN

sub u_seconds {
   my ($t_day) = @_;  # Arguments

   my ($fy, $fm, $fd);
   $t_day =~ /^(\d\d\d\d)(\d\d)(\d\d)/;
   $fy = $1;  $fm = $2;  $fd = $3;
   return Time::Local::timegm(0, 0, 0, $fd, $fm-1, $fy-1900);
#   return Time::Local::timelocal(0, 0, 0, $fd, $fm-1, $fy-1900);
   }

sub modify_date {
   my ($t_day, $d_incr) = @_;  # Arguments

   my $fs = u_seconds($t_day);

   my @fyd = gmtime($fs + $d_incr * 24*60*60);  # + hours
#   my @fyd = localtime($fs + $d_incr * 24*60*60);  # + hours
   my $r = sprintf "%04d%02d%02d",
      $fyd[5]+1900, $fyd[4]+1, $fyd[3];  # yyyy mm dd
   return $r;
   }

sub g_fmt_date {
   my ($t_day, $h_incr) = @_;  # Arguments (h_incr in minutes)

   my $fs = u_seconds($t_day);

   if (defined($h_incr)) { $fs += $h_incr*60; }

   my @fyd = gmtime($fs);
#   my @fyd = localtime($fs);
   my $r = sprintf '"%04d%02d%02d.%02d%02d"',
      $fyd[5]+1900,$fyd[4]+1,$fyd[3], $fyd[2],$fyd[1];  # YYYYmmdd.hhmm
   return $r;
   }

sub fn_date {  # filename from date
   my ($t_day) = @_;  # Arguments

   my ($fy, $fm, $fd);
   $t_day =~ /^(\d\d\d\d)(\d\d)(\d\d)/;
   $fy = $1;  $fm = $2;  $fd = $3;
   return "$fy/$fm/$t_day";
   }

sub log_ts {  # Timestamp string for log messages
    my @fyd = localtime();
    my $r = sprintf '%04d%02d%02d.%02d%02d%02s',  # YYYYmmdd.hhmmss
       $fyd[5]+1900,$fyd[4]+1,$fyd[3], $fyd[2],$fyd[1],$fyd[0];
    return "$r ($$)";
    }


sub get_dist_params {  # From ## records
   my ($from_file, $utc_ref,$st_ref,$dist_ref,$cols_ref) = @_;  # Arguments

   if (defined($from_file)) {
      open(FROM, $from_file)
      or die "Can't open from_file: $!\n";
      }
    
   my (@attribs, @lo, @hi, @nvalues);
   my $n = 0;  my $mycols = 1;

   while (<FROM>) { # read  flows file
      chop;  # Trim new line
      if (/^#/) {  # Starts with a #
         if (/^##NeTraMet /) {
            if (/\(UTC\)/) {  $$utc_ref = 1; }
	     print STDERR " ==== utc_in = 1\n";
            }
         elsif (/^##plt_stat /) {
            my @items = split(/ +/, $_);
            foreach my $item (@items) {
               if ($item =~ /(.+)=(\d+)\.\.(\d+)/) {
                  $attribs[$n] = $1;
	          $lo[$n] = $2;  $hi[$n] = $3;
                  $nvalues[$n] = $3-$2+1;
	          $mycols *= $nvalues[$n];
	          ++$n;
                  }
               elsif ($item =~ /stats=(\d+)/) {
                  $$st_ref = $1;
                  }
               elsif ($item =~ /distribs=(\d+)/) {
                  $$dist_ref = $1;  $mycols *= $1;
                  }
               elsif ($item =~ /cols=(\d+)/) {
                  $$cols_ref = $1;
                  }
               }
            close(FROM);
            return;
	    }
	 } 
      next;
      }
   }

sub get_time {  # From #Time records
      # $c_interval = -c interval from ## header record
      # $t_fmt specifies format of time returned in $x (for gnuplot times)
      # $delta returns delay variation from $last_c_time, seconds
   my ($rec_time, $c_interval,$t_fmt, $x_ref,$delta_ref) = @_;  # Arguments

#. print "get_time($rec_time, $c_interval,$t_fmt):\n";
   if ($rec_time =~ /(^\d+):(\d+):(\d+) +\w+ +(\d+) +(\w+) +(\d+) +/) {
      #    HH    MM   SS    day    dd    mm     yyyy  >>day not saved<<
#. print "   $1 $2 $3 $4 $5 $6 $main::month{$5}\n";
      my $c_time = timegm($3,$2,$1, $4,$main::month{$5}-1,$6-1900);
#      my $c_time = timelocal($3,$2,$1, $4,$main::month{$5}-1,$6-1900);
#. print "   \$c_time=$c_time, \$last_c_time=$last_c_time\n";
      if ($c_interval ne '') {
         if ($last_c_time != 0) {
            my $interval = $c_time - $last_c_time;
            my $delta = $interval - $c_interval;
            if (abs($delta) > 2) {
               print STDERR $_, "\n";
               print STDERR "   interval = $interval <<<<<<<<<\n";
               }
            $$delta_ref = $delta;
            }
         $last_c_time = $c_time;
         }
      if ($t_fmt == 1) {  # 5 integers for R
         $$x_ref = sprintf("%02d %02d %02d %02d %02d",  # mm dd HH MM SS
            $main::month{$5},$4, $1,$2,$3);
         }
      else {  # default, YYYYmmdd.HHMMSS for gnuplot data files (.plt)
         $$x_ref = sprintf("%04d%02d%02d.%02d%02d%02d", 
            $6,$main::month{$5},$4, $1,$2,$3);              # YYYYmmdd.HHMMSS
         }
      }
   }


sub unpack_format_record {
   my ($fmt) = @_;

   %main::format = ();  @main::type = ();  @main::dif  = ();

   my $out_format;
   for ($main::nf = 0; $fmt =~ s/^([(<\[]\w*[)>\]]|\w+\S*)//; ) {
      my $next = $1;  # Split up fmt
      $fmt =~ s/\s*//;
      $out_format = $out_format . " $next";
         # put_flow() writes flows with single spaces between items

      if ($next =~ /^[\w]/ && $next !~ /^\w_/) {  # (<[ and a_ are OK 
         next if $next =~ /[\W]/;  # Other non_alpha not in attrib names
         }

      $main::type[$main::nf] = 0;
      if ($next =~ s/^\((.+)\)$/$1/ ) {
         $main::type[$main::nf] = 2;  # Distribution
         }
      elsif ($next =~ s/^<(.+)>$/$1/ ) {
         $main::type[$main::nf] = 3;  # TCPStreamData
         }
      elsif ($next =~ s/^\[(.+)\]$/$1/ ) {
         $main::type[$main::nf] = 1;  # TCPdata
         }
      $main::format{$next} = $main::nf;
#. print "fmt: attrib $next, index=$main::nf, type=$main::type[$main::nf]\n";
      $main::dif[$main::nf] = 0;
      if ($next =~ /^d_/) {
         $main::dif[$main::nf] = 1;  # Delta attribute
         }
      ++$main::nf;
      }
   return $out_format;
   }

sub get_attrib_values {  # From command line
                         # Returns $item whoich stopped parse
   my ($errs_ref) = @_;  # Nbr of errors

   my $item;
   for (my $a = 0; $item = shift @ARGV; ) {  # Grab items from command line
#. print "AV: \$item=$item\n";
      last if $item eq '+';  # Stop when we find a '+'
      last if $item eq '%';  #    or a '%'
      last if $item eq '/';  #    or a '/'
      $item = lc($item);

      if ($item =~ /^\^(\d+)/) {  # Caret-introduced int parameter value
         $main::caret = $1;
         next;
         }
      elsif ($item =~ s/^\_//) {  # Underscore-introduced string param value
         $main::uscore = $item;
         next;
         }
      elsif ($item =~ s/^\-//) {  # Minus-introduced string parameter value
         $main::minus = $item;
         next;
         }
      elsif ($item =~ s/^\@//) {  # At-introduced string parameter value
         $main::at = $item;
         next;
         }
      elsif ($item =~ s/^\!//) {  # At-introduced string parameter value
         $main::shriek = $item;
         next;
         }
      elsif ($item =~ s/^\?//) {  # At-introduced string parameter value
         $main::query = $item;
         next;
         }

      my ($atr, $test, $val, $str_atr, $valh, $range);
#.print "-1- item=$item\n";
      if ($item =~ s/\.\.(\d+)//) { 
         $valh = $1;
         $range = '..';  # Range
         }
      elsif ($item =~ s/\,(.+)$//) {
         $valh = $1;
         $range = ',';  # List
         }
      else { 
         $valh = '';  $range = '';
         }
#.print "-2- valh-$valh, range=>$range<\n";
      if ($item =~ /^(\w+)([=!\#]+)([\w.]+)/) {
         $atr = $1;  $test = substr($2,0,1);  $val = $3;

#         if ($range eq '..') {
#            $c1 = $val;  $c2 = $valh;
#            }
#         elsif ($range eq ',') {
#            $values{$val} = $c1 = $c2 = 1;
#            for (my $vh = $valh . ',';  $vh =~ s/^(\d+)\,//; ) {
#               $values{$1} = ++$c2;
#               }
#            }
#         else {  # 0 or 1 value specified
#            $c1 = $c2 = $val;
#            }

         $str_atr = $val =~ /\./ ;  # Dots appear in IP addresses
         if ($valh ne '' && $val > $valh) {
            print STDERR "Bad attribute range for $atr ($val > $valh)\n";
            ++$$errs_ref;
            }
         if ($valh eq '' && !$str_atr) { $valh = $val; }
         }
      else {  # No values, all flows match.  Use attrib values as key.
         $atr = $item;  $test = '=';  $val = '';
         }

      if (defined $main::format{$atr} ) {
         $main::attrb[$a] = $atr;  $main::attst[$a] = $test;
         $main::attvl[$a] = $val;  $main::att_str[$a] = $str_atr;
         $main::attvh[$a] = $valh;
         $main::nvalues[$a] =
            $valh ne '' ? $valh-$val+1 : ($val ne '' ? 1 : 0);
         printf STDERR " %s %s %s %s (%d values)\n", $atr, $test, $val,
            $valh ne '' ? ".. $valh" : '', $main::nvalues[$a];
         ++$a;
         }
      else {
         print STDERR "Unknown attribute '$atr'\n";
         ++$$errs_ref;
         }
      }

   return $item;
   }

sub get_flow{  # Take $_, return ref to array of values

   my @value = ();
   for (my $i = 0; s/(^[^0123456789ABCDEF(>\[]*)//; ) {
         # Remove leading non-numerics followed by (, <, [ or a value
         # NOTE: .- appear in addresses, : appears in V6 addresses

      last if !( s/^([(<\]]|.[0123456789ABCDEF.-:]*)// );
         # Get next likely data value

      if ($1 =~ /^[ABCDEF]$/) {  # Single hex letter
         # Assume it's part of a separator, e.g. " From: "
         s/^[^0123456789ABCDEF(<\]]*//;
         next;
         }

      if ($1 eq '(') {  #  have distrib
	 s/\W*([^\)]+)\)//;  # remove everything upto next ) save it in $1

	 my @v = split(/ +/, $1);   # put the distribution values in @v
	 my $v = {};                # $v points to a hash
	 if (($v->{Trans} = $v[0]) != 0 ) {  # construct structure
	    $v->{Lower} = $v[2];
	    $v->{Upper} = $v[3];
	    if ($v[1] > 0x0F) {  # Upper scale delta
	       $v->{Scale} = $v[1] & 0x0F;
               $v->{Upper} *= 10**($v[1] >> 4);
 	       }
	    else { $v->{Scale} = $v[1]; }
	    $v->{Buckets} = $v[4];
	    $v->{P1} = $v[5];
	    $v->{P2} = $v[6];
	    $v->{P3} = $v[7];
	    $v->{Dist} = [splice(@v,8)]; # copies from element 8 to end
#            if ($v->{P1} != 0) {
#               $samples_per_int = $c_interval/$v->{P1};
#               $v->{Dist}->[0] %= $samples_per_int;
#                  # Ignore empty collection intervals
#               }
	    }
         $value[$i] = $v;
	 } 
      elsif ($1 eq '<') {  # TCPStreamData
	 s/([^>]+)\>//;  # Remove everything upto next >, save it in $1 
         my @v = split(/ +/, $1);
	 $value[$i] = \@v;
         }
      elsif ($1 eq '[') {  # TCPdata
	 s/([^\]]+)\]//;  # Remove everything upto next ], save it in $1 
         my @v = split(/ +/, $1);
	 $value[$i] = \@v;
         }
      else {
	 $value[$i] = $1;
	 }

      ++$i;
      }

   return \@value;  # Return ref to array
   }

sub match_flow{  # Returns flow key, '' if match fails
   my ($vref) = @_;

   my ($av, $match, $fail, $key);
   $fail = 0;  $key = '';
   for (my $q = 0; $q != scalar(@main::attrb); ++$q) {
      $av = $vref->[$main::format{$main::attrb[$q]}];
      $key = $key . ($main::att_str[$q] ? $av : sprintf("%03d ", $av));
      if ($main::attvl[$q] ne '') {
         if ($main::attvh[$q] eq '') {
            $match = $main::att_str[$q] ?
               $av eq $main::attvl[$q] : $av == $main::attvl[$q];
            }
         else {
            $match = ($av >= $main::attvl[$q] && $av <= $main::attvh[$q]);
            }
         if ($match) {
            last if $fail = $main::attst[$q] eq '!';
            }
         else {
            last if $fail = $main::attst[$q] eq '=';
            }
         } 
      }
   return $fail ? '' : $key;
   }

sub put_flow{  # Write output record
   my ($FILE, $vref) = @_;

   for (my $j = 0; $j != $main::nf; ++$j) {
#.print "=== j=$j, type=$main::type[$j]\n";
      if ($main::type[$j] == 0) {  # Scalar
         print $FILE "$vref->[$j] ";
         }
      elsif ($main::type[$j] == 2) {  # Distribution
         my $hr = $vref->[$j];
         if (!defined($hr->{Trans}) || $hr->{Trans} == 0) {
            print $FILE "(0) ";  # Empty distribution
	    }
	 else {
            print $FILE 
               "($hr->{Trans} $hr->{Scale} $hr->{Lower} $hr->{Upper}  ";
            print $FILE 
               "$hr->{Buckets} $hr->{P1} $hr->{P2} $hr->{P3}  ";
            my @bucket = @{ $hr->{Dist} };
            print $FILE "@bucket) ";
            }
         }
      else {  # TCPdata
         my @data = @{ $vref->[$j] };
         print $FILE "[@data] ";
         }
      }
   print $FILE "\n";
   }


return 1;  # Perl seems to need this when loading the package


# $Log: rt_dist.pm,v $
# Revision 1.2  2006/03/27 02:53:29  nevil
# Updates to 51b16, on way home from IETF65
#
# Revision 1.6  2006/03/22 16:39:13  nevil
# 51b16
